#ifndef SAM_H
#define SAM_H

#ifdef __cplusplus
extern "C"
{
#endif

    void SetInput(char *_input);
    void SetSpeed(unsigned char _speed);
    void SetPitch(unsigned char _pitch);
    void SetMouth(unsigned char _mouth);
    void SetThroat(unsigned char _throat);
    void EnableSingmode();
    void DisableSingmode();
    void EnableDebug();

    int SAMMain();

    char *GetBuffer();
    int GetBufferLength();
    void FreeBuffer();
    
    //char input[]={"/HAALAOAO MAYN NAAMAEAE IHSTT SAEBAASTTIHAAN \x9b\x9b\0"};
    //unsigned char input[]={"/HAALAOAO \x9b\0"};
    //unsigned char input[]={"AA \x9b\0"};
    //unsigned char input[] = {"GUH5DEHN TAEG\x9b\0"};

    //unsigned char input[]={"AY5 AEM EY TAO4LXKIHNX KAX4MPYUX4TAH. GOW4 AH/HEH3D PAHNK.MEYK MAY8 DEY.\x9b\0"};
    //unsigned char input[]={"/HEH3LOW2, /HAW AH YUX2 TUXDEY. AY /HOH3P YUX AH FIYLIHNX OW4 KEY.\x9b\0"};
    //unsigned char input[]={"/HEY2, DHIHS IH3Z GREY2T. /HAH /HAH /HAH.AYL BIY5 BAEK.\x9b\0"};
    //unsigned char input[]={"/HAH /HAH /HAH \x9b\0"};
    //unsigned char input[]={"/HAH /HAH /HAH.\x9b\0"};
    //unsigned char input[]={".TUW BIY5Y3,, OHR NAA3T - TUW BIY5IYIY., DHAE4T IHZ DHAH KWEH4SCHAHN.\x9b\0"};
    //unsigned char input[]={"/HEY2, DHIHS \x9b\0"};

    //unsigned char input[]={" IYIHEHAEAAAHAOOHUHUXERAXIX  \x9b\0"};
    //unsigned char input[]={" RLWWYMNNXBDGJZZHVDH \x9b\0"};
    //unsigned char input[]={" SSHFTHPTKCH/H \x9b\0"};

    //unsigned char input[]={" EYAYOYAWOWUW ULUMUNQ YXWXRXLX/XDX\x9b\0"};

#ifdef __cplusplus
}
#endif

#endif
